# SPDX-License-Identifier: Apache-2.0
# Defines a number of default file names, column names and other strings
VIRTUAL_FIBERS_FN = "virtual-fibers.csv"

XYZ = ["x", "y", "z"]
UVW = ["u", "v", "w"]
FLAT_COORDINATES = ["flat_x", "flat_y"]
SS_COORDINATES = ["ss_flat_x", "depth", "ss_flat_y"]
GID = "node_ids"
FIBER_GID = "sgid"
PROJECTION = "projection"
DEFAULT_EDGES = "default"
DEFAULT_PIXEL_SZ = 34.0
